@@ -1,5 +1,6 @@
 /*
- *  Copyright (c) 2008 Mike Massonnet <mmassonnet@xfce.org>
+ *  Copyright (c) 2008-2009 Mike Massonnet <mmassonnet@xfce.org>
+ *  Copyright (c) 2009 Vincent Legout <vincent@xfce.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -9,11 +10,11 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -24,11 +25,9 @@
 #include <libmpd/libmpd.h>
 
 #include "mpdclient.h"
+#include "preferences.h"
 
-#define MPD_HOST "localhost"
-#define MPD_PORT 6600
-
-#define XFMPC_MPDCLIENT_GET_PRIVATE(o) \
+#define GET_PRIVATE(o) \
     (G_TYPE_INSTANCE_GET_PRIVATE ((o), XFMPC_TYPE_MPDCLIENT, XfmpcMpdclientPrivate))
 
 
@@ -48,7 +47,7 @@ enum
   LAST_SIGNAL
 };
 
-static guint            xfmpc_mpdclient_signals[LAST_SIGNAL] = { 0 };
+static guint            signals[LAST_SIGNAL] = { 0 };
 
 
 
@@ -58,9 +57,14 @@ static void             xfmpc_mpdclient_finalize                (GObject *object
 
 static void             xfmpc_mpdclient_initenv                 (XfmpcMpdclient *mpdclient);
 
-static void             cb_xfmpc_mpdclient_status_changed       (MpdObj *mi,
+static void             cb_status_changed                       (MpdObj *mi,
                                                                  ChangedStatusType what,
                                                                  gpointer user_data);
+static gchar *          _get_formatted_name                     (mpd_Song *song);
+static gchar *          _get_formatted_name_predefined          (mpd_Song *song,
+                                                                 XfmpcSongFormat song_format);
+static gchar *          _get_formatted_name_custom              (mpd_Song *song,
+                                                                 const gchar *format);
 
 
 
@@ -83,6 +87,7 @@ struct _XfmpcMpdclientClass
 struct _XfmpcMpdclient
 {
   GObject                   parent;
+  /*<private>*/
   XfmpcMpdclientPrivate    *priv;
 };
 
@@ -92,6 +97,7 @@ struct _XfmpcMpdclientPrivate
   gchar                    *host;
   guint                     port;
   gchar                    *passwd;
+  gboolean                  env_cached;
 };
 
 
@@ -140,7 +146,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
   gobject_class = G_OBJECT_CLASS (klass);
   gobject_class->finalize = xfmpc_mpdclient_finalize;
 
-  xfmpc_mpdclient_signals[SIG_CONNECTED] =
+  signals[SIG_CONNECTED] =
     g_signal_new ("connected", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, connected),
@@ -148,7 +154,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
 
-  xfmpc_mpdclient_signals[SIG_SONG_CHANGED] =
+  signals[SIG_SONG_CHANGED] =
     g_signal_new ("song-changed", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, song_changed),
@@ -156,7 +162,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
 
-  xfmpc_mpdclient_signals[SIG_PP_CHANGED] =
+  signals[SIG_PP_CHANGED] =
     g_signal_new ("pp-changed", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, pp_changed),
@@ -165,7 +171,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   G_TYPE_NONE, 1,
                   G_TYPE_BOOLEAN);
 
-  xfmpc_mpdclient_signals[SIG_TIME_CHANGED] =
+  signals[SIG_TIME_CHANGED] =
     g_signal_new ("time-changed", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, time_changed),
@@ -175,7 +181,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   G_TYPE_INT,
                   G_TYPE_INT);
 
-  xfmpc_mpdclient_signals[SIG_VOLUME_CHANGED] =
+  signals[SIG_VOLUME_CHANGED] =
     g_signal_new ("volume-changed", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, volume_changed),
@@ -184,7 +190,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   G_TYPE_NONE, 1,
                   G_TYPE_INT);
 
-  xfmpc_mpdclient_signals[SIG_STOPPED] =
+  signals[SIG_STOPPED] =
     g_signal_new ("stopped", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, stopped),
@@ -192,7 +198,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
 
-  xfmpc_mpdclient_signals[SIG_DATABASE_CHANGED] =
+  signals[SIG_DATABASE_CHANGED] =
     g_signal_new ("database-changed", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, database_changed),
@@ -200,7 +206,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
 
-  xfmpc_mpdclient_signals[SIG_PLAYLIST_CHANGED] =
+  signals[SIG_PLAYLIST_CHANGED] =
     g_signal_new ("playlist-changed", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, playlist_changed),
@@ -208,7 +214,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
 
-  xfmpc_mpdclient_signals[SIG_REPEAT] =
+  signals[SIG_REPEAT] =
     g_signal_new ("repeat", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, repeat),
@@ -217,7 +223,7 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
                   G_TYPE_NONE, 1,
                   G_TYPE_BOOLEAN);
 
-  xfmpc_mpdclient_signals[SIG_RANDOM] =
+  signals[SIG_RANDOM] =
     g_signal_new ("random", G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST|G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (XfmpcMpdclientClass, random),
@@ -230,18 +236,16 @@ xfmpc_mpdclient_class_init (XfmpcMpdclientClass *klass)
 static void
 xfmpc_mpdclient_init (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
-  xfmpc_mpdclient_initenv (mpdclient);
-  priv->mi = mpd_new (priv->host, priv->port, priv->passwd);
-  mpd_signal_connect_status_changed (priv->mi, (StatusChangedCallback)cb_xfmpc_mpdclient_status_changed, mpdclient);
+  XfmpcMpdclientPrivate *priv = mpdclient->priv = GET_PRIVATE (mpdclient);
+  priv->mi = mpd_new_default ();
+  mpd_signal_connect_status_changed (priv->mi, (StatusChangedCallback)cb_status_changed, mpdclient);
 }
 
 static void
 xfmpc_mpdclient_finalize (GObject *object)
 {
   XfmpcMpdclient *mpdclient = XFMPC_MPDCLIENT (object);
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   mpd_free (priv->mi);
   (*G_OBJECT_CLASS (parent_class)->finalize) (object);
@@ -268,44 +272,72 @@ xfmpc_mpdclient_get ()
 static void
 xfmpc_mpdclient_initenv (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
+  XfmpcPreferences *preferences = xfmpc_preferences_get ();
+  gboolean use_defaults;
 
-  /* Hostname */
-  priv->host = (g_getenv ("MPD_HOST") != NULL) ?
-      g_strdup (g_getenv ("MPD_HOST")) :
-      g_strdup (MPD_HOST);
+  g_object_get (preferences, "mpd-use-defaults", &use_defaults, NULL);
 
-  /* Port */
-  priv->port = (g_getenv ("MPD_PORT") != NULL) ?
-      (gint) g_ascii_strtoll (g_getenv ("MPD_PORT"), NULL, 0) :
-      MPD_PORT;
+  if (use_defaults)
+    {
+      if (!priv->env_cached)
+        {
+          g_free (priv->host);
+          g_free (priv->passwd);
+          priv->env_cached = TRUE;
 
-  /* Check for password */
-  priv->passwd = NULL;
-  gchar **split = g_strsplit (priv->host, "@", 2);
-  if (g_strv_length (split) == 2)
+          priv->host = (g_getenv ("MPD_HOST") != NULL) ?
+            g_strdup (g_getenv ("MPD_HOST")) :
+            g_strdup ("127.0.0.1");
+
+          priv->port = (g_getenv ("MPD_PORT") != NULL) ?
+            (gint) g_ascii_strtoll (g_getenv ("MPD_PORT"), NULL, 0) :
+            6600;
+
+          priv->passwd = NULL;
+          gchar **split = g_strsplit (priv->host, "@", 2);
+          if (g_strv_length (split) == 2)
+            {
+              g_free (priv->host);
+              priv->host = g_strdup (split[1]);
+              priv->passwd = g_strdup (split[0]);
+            }
+          g_strfreev (split);
+        }
+    }
+  else
     {
       g_free (priv->host);
-      priv->host = g_strdup (split[0]);
-      priv->passwd = g_strdup (split[1]);
+      g_free (priv->passwd);
+      priv->env_cached = FALSE;
+
+      g_object_get (preferences,
+                    "mpd-hostname", &priv->host,
+                    "mpd-port", &priv->port,
+                    "mpd-password", &priv->passwd,
+                    NULL);
     }
-  g_strfreev (split);
 }
 
 gboolean
 xfmpc_mpdclient_connect (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (xfmpc_mpdclient_is_connected (mpdclient))
     return TRUE;
 
+  xfmpc_mpdclient_initenv (mpdclient);
+  mpd_set_hostname (priv->mi, priv->host);
+  mpd_set_port (priv->mi, priv->port);
+  mpd_set_password (priv->mi, (priv->passwd != NULL) ? priv->passwd : "");
+
   if (mpd_connect (priv->mi) != MPD_OK)
     return FALSE;
 
   mpd_send_password (priv->mi);
 
-  g_signal_emit_by_name (mpdclient, "connected");
+  g_signal_emit (mpdclient, signals[SIG_CONNECTED], 0);
 
   return TRUE;
 }
@@ -313,7 +345,7 @@ xfmpc_mpdclient_connect (XfmpcMpdclient *mpdclient)
 void
 xfmpc_mpdclient_disconnect (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (xfmpc_mpdclient_is_connected (mpdclient))
     mpd_disconnect (priv->mi);
@@ -322,7 +354,7 @@ xfmpc_mpdclient_disconnect (XfmpcMpdclient *mpdclient)
 gboolean
 xfmpc_mpdclient_is_connected (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   return mpd_check_connected (priv->mi);
 }
@@ -330,7 +362,7 @@ xfmpc_mpdclient_is_connected (XfmpcMpdclient *mpdclient)
 gboolean
 xfmpc_mpdclient_previous (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_player_prev (priv->mi) != MPD_OK)
     return FALSE;
@@ -350,7 +382,7 @@ xfmpc_mpdclient_pp (XfmpcMpdclient *mpdclient)
 gboolean
 xfmpc_mpdclient_play (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_player_play (priv->mi) != MPD_OK)
     return FALSE;
@@ -361,7 +393,7 @@ xfmpc_mpdclient_play (XfmpcMpdclient *mpdclient)
 gboolean
 xfmpc_mpdclient_pause (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_player_pause (priv->mi) != MPD_OK)
     return FALSE;
@@ -372,7 +404,7 @@ xfmpc_mpdclient_pause (XfmpcMpdclient *mpdclient)
 gboolean
 xfmpc_mpdclient_stop (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_player_stop (priv->mi) != MPD_OK)
     return FALSE;
@@ -383,7 +415,7 @@ xfmpc_mpdclient_stop (XfmpcMpdclient *mpdclient)
 gboolean
 xfmpc_mpdclient_next (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_player_next (priv->mi) != MPD_OK)
     return FALSE;
@@ -395,7 +427,7 @@ gboolean
 xfmpc_mpdclient_set_id (XfmpcMpdclient *mpdclient,
                         gint id)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   return mpd_player_play_id (priv->mi, id) == MPD_OK ? TRUE : FALSE;
 }
@@ -404,7 +436,7 @@ gboolean
 xfmpc_mpdclient_set_song_time (XfmpcMpdclient *mpdclient,
                                guint time)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_player_seek (priv->mi, time) != MPD_OK)
     return FALSE;
@@ -416,7 +448,7 @@ gboolean
 xfmpc_mpdclient_set_volume (XfmpcMpdclient *mpdclient,
                             guint8 volume)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_status_set_volume (priv->mi, volume) < 0)
     return FALSE;
@@ -428,7 +460,7 @@ gboolean
 xfmpc_mpdclient_set_repeat (XfmpcMpdclient *mpdclient,
                             gboolean repeat)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (MPD_OK != mpd_player_set_repeat (priv->mi, repeat))
     return FALSE;
@@ -440,7 +472,7 @@ gboolean
 xfmpc_mpdclient_set_random (XfmpcMpdclient *mpdclient,
                             gboolean random)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (MPD_OK != mpd_player_set_random (priv->mi, random))
     return FALSE;
@@ -453,7 +485,7 @@ xfmpc_mpdclient_set_random (XfmpcMpdclient *mpdclient,
 gint
 xfmpc_mpdclient_get_pos (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   gint pos = mpd_player_get_current_song_pos (priv->mi);
   if (pos < 0)
@@ -465,7 +497,7 @@ xfmpc_mpdclient_get_pos (XfmpcMpdclient *mpdclient)
 gint
 xfmpc_mpdclient_get_id (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   return mpd_player_get_current_song_id (priv->mi);
 }
@@ -473,7 +505,7 @@ xfmpc_mpdclient_get_id (XfmpcMpdclient *mpdclient)
 const gchar *
 xfmpc_mpdclient_get_artist (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   mpd_Song *song = mpd_playlist_get_current_song (priv->mi);
   if (G_UNLIKELY (NULL == song))
@@ -488,7 +520,7 @@ xfmpc_mpdclient_get_artist (XfmpcMpdclient *mpdclient)
 const gchar *
 xfmpc_mpdclient_get_title (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   mpd_Song *song = mpd_playlist_get_current_song (priv->mi);
   if (G_UNLIKELY (NULL == song))
@@ -503,7 +535,7 @@ xfmpc_mpdclient_get_title (XfmpcMpdclient *mpdclient)
 const gchar *
 xfmpc_mpdclient_get_album (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   mpd_Song *song = mpd_playlist_get_current_song (priv->mi);
   if (G_UNLIKELY (NULL == song))
@@ -518,7 +550,7 @@ xfmpc_mpdclient_get_album (XfmpcMpdclient *mpdclient)
 const gchar *
 xfmpc_mpdclient_get_date (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   mpd_Song *song = mpd_playlist_get_current_song (priv->mi);
   if (G_UNLIKELY (NULL == song))
@@ -533,24 +565,21 @@ xfmpc_mpdclient_get_date (XfmpcMpdclient *mpdclient)
 gint
 xfmpc_mpdclient_get_time (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   return mpd_status_get_elapsed_song_time (priv->mi);
 }
 
 gint
 xfmpc_mpdclient_get_total_time (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   return mpd_status_get_total_song_time (priv->mi);
 }
 
 guint8
 xfmpc_mpdclient_get_volume (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   gint volume = mpd_status_get_volume (priv->mi);
   return (volume < 0) ? 100 : volume;
 }
@@ -558,31 +587,28 @@ xfmpc_mpdclient_get_volume (XfmpcMpdclient *mpdclient)
 gboolean
 xfmpc_mpdclient_get_repeat (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   return mpd_player_get_repeat (priv->mi);
 }
 
 gboolean
 xfmpc_mpdclient_get_random (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   return mpd_player_get_random (priv->mi);
 }
+
 gboolean
 xfmpc_mpdclient_is_playing (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   return mpd_player_get_state (priv->mi) == MPD_PLAYER_PLAY;
 }
 
 gboolean
 xfmpc_mpdclient_is_stopped (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   return mpd_player_get_state (priv->mi) == MPD_PLAYER_STOP;
 }
 
@@ -591,54 +617,53 @@ xfmpc_mpdclient_is_stopped (XfmpcMpdclient *mpdclient)
 void
 xfmpc_mpdclient_update_status (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
-
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
   mpd_status_update (priv->mi);
 }
 
 static void
-cb_xfmpc_mpdclient_status_changed (MpdObj *mi,
-                                   ChangedStatusType what,
-                                   gpointer user_data)
+cb_status_changed (MpdObj *mi,
+                   ChangedStatusType what,
+                   gpointer user_data)
 {
   XfmpcMpdclient *mpdclient = XFMPC_MPDCLIENT (user_data);
   g_return_if_fail (G_LIKELY (NULL != user_data));
 
   if (what & MPD_CST_DATABASE)
-    g_signal_emit_by_name (mpdclient, "database-changed");
+    g_signal_emit (mpdclient, signals[SIG_DATABASE_CHANGED], 0);
 
   if (what & MPD_CST_PLAYLIST)
-    g_signal_emit_by_name (mpdclient, "playlist-changed");
+    g_signal_emit (mpdclient, signals[SIG_PLAYLIST_CHANGED], 0);
 
-  if (what & MPD_CST_SONGID)
-    g_signal_emit_by_name (mpdclient, "song-changed");
+  if (what & MPD_CST_SONGID && !(what & MPD_CST_DATABASE) && !(what & MPD_CST_PLAYLIST))
+    g_signal_emit (mpdclient, signals[SIG_SONG_CHANGED], 0);
 
   if (what & MPD_CST_STATE)
     {
       gint state = mpd_player_get_state (mi);
       if (state == MPD_PLAYER_STOP)
-        g_signal_emit_by_name (mpdclient, "stopped");
+        g_signal_emit (mpdclient, signals[SIG_STOPPED], 0);
       else if (state == MPD_PLAYER_PLAY || state == MPD_PLAYER_PAUSE)
-        g_signal_emit_by_name (mpdclient, "pp-changed",
-                               state == MPD_PLAYER_PLAY);
+        g_signal_emit (mpdclient, signals[SIG_PP_CHANGED], 0,
+                       state == MPD_PLAYER_PLAY);
     }
 
   if (what & MPD_CST_VOLUME)
-    g_signal_emit_by_name (mpdclient, "volume-changed",
-                           xfmpc_mpdclient_get_volume (mpdclient));
+    g_signal_emit (mpdclient, signals[SIG_VOLUME_CHANGED], 0,
+                   xfmpc_mpdclient_get_volume (mpdclient));
 
   if (what & (MPD_CST_ELAPSED_TIME|MPD_CST_TOTAL_TIME))
-    g_signal_emit_by_name (mpdclient, "time-changed",
-                           xfmpc_mpdclient_get_time (mpdclient),
-                           xfmpc_mpdclient_get_total_time (mpdclient));
+    g_signal_emit (mpdclient, signals[SIG_TIME_CHANGED], 0,
+                   xfmpc_mpdclient_get_time (mpdclient),
+                   xfmpc_mpdclient_get_total_time (mpdclient));
 
   if (what & MPD_CST_REPEAT)
-    g_signal_emit_by_name (mpdclient, "repeat",
-                           xfmpc_mpdclient_get_repeat (mpdclient));
+    g_signal_emit (mpdclient, signals[SIG_REPEAT], 0,
+                   xfmpc_mpdclient_get_repeat (mpdclient));
 
   if (what & MPD_CST_RANDOM)
-    g_signal_emit_by_name (mpdclient, "random",
-                           xfmpc_mpdclient_get_random (mpdclient));
+    g_signal_emit (mpdclient, signals[SIG_RANDOM], 0,
+                   xfmpc_mpdclient_get_random (mpdclient));
 }
 
 
@@ -646,7 +671,7 @@ cb_xfmpc_mpdclient_status_changed (MpdObj *mi,
 gboolean
 xfmpc_mpdclient_queue_commit (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_playlist_queue_commit (priv->mi) != MPD_OK)
     return FALSE;
@@ -658,7 +683,7 @@ gboolean
 xfmpc_mpdclient_queue_add (XfmpcMpdclient *mpdclient,
                            const gchar *path)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_playlist_queue_add (priv->mi, (gchar *)path) != MPD_OK)
     return FALSE;
@@ -670,7 +695,7 @@ gboolean
 xfmpc_mpdclient_queue_remove_id (XfmpcMpdclient *mpdclient,
                                  gint id)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_playlist_queue_delete_id (priv->mi, id) != MPD_OK)
     return FALSE;
@@ -678,16 +703,32 @@ xfmpc_mpdclient_queue_remove_id (XfmpcMpdclient *mpdclient,
   return TRUE;
 }
 
+gboolean
+xfmpc_mpdclient_queue_clear (XfmpcMpdclient *mpdclient)
+{
+  gint id;
+
+  while (xfmpc_mpdclient_playlist_read (mpdclient, &id, NULL, NULL, NULL, NULL))
+    {
+      if (!xfmpc_mpdclient_queue_remove_id (mpdclient, id))
+        return FALSE;
+    }
+
+  return TRUE;
+}
+
 
 
 gboolean
 xfmpc_mpdclient_playlist_read (XfmpcMpdclient *mpdclient,
                                gint *id,
+                               gint *pos,
+                               gchar **filename,
                                gchar **song,
                                gchar **length)
 {
   static MpdData       *data = NULL;
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (NULL == data)
     data = mpd_playlist_get_changes (priv->mi, -1);
@@ -696,17 +737,14 @@ xfmpc_mpdclient_playlist_read (XfmpcMpdclient *mpdclient,
 
   if (NULL != data)
     {
-      if (NULL != data->song->title)
-        {
-          if (data->song->artist)
-            *song = g_strdup_printf ("%s - %s", data->song->artist, data->song->title);
-          else
-            *song = g_strdup (data->song->title);
-        }
-      else
-        *song = g_path_get_basename (data->song->file);
-
-      *length = g_strdup_printf ("%d:%02d", data->song->time / 60, data->song->time % 60);
+      if (NULL != filename)
+        *filename = g_strdup (data->song->file);
+      if (NULL != song)
+        *song = _get_formatted_name (data->song);
+      if (NULL != length)
+        *length = g_strdup_printf ("%d:%02d", data->song->time / 60, data->song->time % 60);
+      if (NULL != pos)
+      	*pos = data->song->pos;
       *id = data->song->id;
     }
 
@@ -716,7 +754,7 @@ xfmpc_mpdclient_playlist_read (XfmpcMpdclient *mpdclient,
 gboolean
 xfmpc_mpdclient_playlist_clear (XfmpcMpdclient *mpdclient)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
 
   if (mpd_playlist_clear (priv->mi) != FALSE)
     return FALSE;
@@ -724,6 +762,79 @@ xfmpc_mpdclient_playlist_clear (XfmpcMpdclient *mpdclient)
   return TRUE;
 }
 
+gint
+xfmpc_mpdclient_playlist_get_length (XfmpcMpdclient *mpdclient)
+{
+  return mpd_playlist_get_playlist_length (mpdclient->priv->mi);
+}
+
+gint
+xfmpc_mpdclient_playlist_get_total_time (XfmpcMpdclient *mpdclient)
+{
+  XfmpcMpdclientPrivate *priv = mpdclient->priv;
+  MpdData               *data;
+  gint                   seconds;
+
+  for (seconds = 0, data = mpd_playlist_get_changes (priv->mi, -1);
+       data != NULL; data = mpd_data_get_next (data))
+    seconds += data->song->time;
+
+  return seconds;
+}
+
+gboolean
+xfmpc_mpdclient_playlist_has_filename (XfmpcMpdclient *mpdclient,
+                                       const gchar *filename,
+                                       gboolean is_dir)
+{
+  XfmpcMpdclientPrivate *priv = mpdclient->priv;
+  MpdData               *data;
+  gchar                 *basename;
+  gint                   id;
+  gboolean               found = FALSE;
+
+  if (is_dir && strcmp (filename, ""))
+    {
+      while (xfmpc_mpdclient_playlist_read (mpdclient, &id, NULL, &basename, NULL, NULL))
+        {
+          if (g_str_has_prefix (basename, filename))
+            found = TRUE;
+
+          g_free (basename);
+        }
+
+        if (found)
+          return TRUE;
+    }
+  else
+    {
+      mpd_playlist_search_start (priv->mi, TRUE);
+      mpd_playlist_search_add_constraint (priv->mi, MPD_TAG_ITEM_FILENAME, filename);
+      data = mpd_playlist_search_commit (priv->mi);
+
+      if (data != NULL)
+        {
+          mpd_data_free (data);
+          return TRUE;
+        }
+    }
+
+  return FALSE;
+}
+
+
+
+gboolean
+xfmpc_mpdclient_database_refresh (XfmpcMpdclient *mpdclient)
+{
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
+
+  if (mpd_database_update_dir (priv->mi, "/") != MPD_OK)
+    return FALSE;
+
+  return TRUE;
+}
+
 gboolean
 xfmpc_mpdclient_database_read (XfmpcMpdclient *mpdclient,
                                const gchar *dir,
@@ -731,8 +842,8 @@ xfmpc_mpdclient_database_read (XfmpcMpdclient *mpdclient,
                                gchar **basename,
                                gboolean *is_dir)
 {
-  static MpdData       *data = NULL;
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
+  static MpdData        *data = NULL;
 
   if (NULL == data)
     data = mpd_database_get_directory (priv->mi, (gchar *)dir);
@@ -752,17 +863,7 @@ xfmpc_mpdclient_database_read (XfmpcMpdclient *mpdclient,
         case MPD_DATA_TYPE_SONG:
           *is_dir = FALSE;
           *filename = g_strdup (data->song->file);
-
-          if (NULL != data->song->title)
-            {
-              if (data->song->artist)
-                *basename = g_strdup_printf ("%s - %s", data->song->artist, data->song->title);
-              else
-                *basename = g_strdup (data->song->title);
-            }
-          else
-            *basename = g_path_get_basename (data->song->file);
-
+          *basename = _get_formatted_name (data->song);
           break;
 
         case MPD_DATA_TYPE_PLAYLIST:
@@ -783,13 +884,281 @@ xfmpc_mpdclient_database_read (XfmpcMpdclient *mpdclient,
 }
 
 gboolean
-xfmpc_mpdclient_database_refresh (XfmpcMpdclient *mpdclient)
+xfmpc_mpdclient_database_search (XfmpcMpdclient *mpdclient,
+                                 const gchar *query,
+                                 gchar **filename,
+                                 gchar **basename)
 {
-  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT_GET_PRIVATE (mpdclient);
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
+  static MpdData        *data = NULL;
+  gchar                **queries;
+  gint                   i;
 
-  if (mpd_database_update_dir (priv->mi, "/") != MPD_OK)
-    return FALSE;
+  if (NULL == data)
+    {
+      queries = g_strsplit (query, " ", -1);
 
-  return TRUE;
+      mpd_database_search_start (priv->mi, FALSE);
+      for (i = 0; queries[i] != NULL; i++)
+        mpd_database_search_add_constraint (priv->mi, MPD_TAG_ITEM_ANY, queries[i]);
+      data = mpd_database_search_commit (priv->mi);
+
+      g_strfreev (queries);
+    }
+  else
+    data = mpd_data_get_next (data);
+
+  if (NULL != data)
+    {
+      *filename = g_strdup (data->song->file);
+      *basename = _get_formatted_name (data->song);
+    }
+
+  return NULL != data;
+}
+
+
+
+static gchar *
+_get_formatted_name (mpd_Song *song)
+{
+  XfmpcPreferences *preferences = xfmpc_preferences_get ();
+  XfmpcSongFormat song_format;
+  gchar *format_custom;
+  gchar *formatted_name;
+
+  g_object_get (preferences,
+                "song-format", &song_format,
+                "song-format-custom", &format_custom,
+                NULL);
+
+  if (NULL == song->title)
+    {
+      formatted_name = g_path_get_basename (song->file);
+    }
+  else if (song_format == XFMPC_SONG_FORMAT_CUSTOM)
+    {
+      formatted_name = _get_formatted_name_custom (song, format_custom);
+    }
+  else
+    {
+      formatted_name = _get_formatted_name_predefined (song, song_format);
+    }
+
+  g_free (format_custom);
+  return formatted_name;
+}
+
+static gchar *
+_get_formatted_name_predefined (mpd_Song *song,
+                                XfmpcSongFormat song_format)
+{
+  gchar *formatted_name, *tmp;
+
+  switch (song_format)
+    {
+    case XFMPC_SONG_FORMAT_TITLE:
+      formatted_name = g_strdup_printf ("%s", song->title);
+      break;
+
+    case XFMPC_SONG_FORMAT_ARTIST_TITLE:
+      if (NULL != song->artist)
+        {
+          formatted_name = g_strdup_printf ("%s - %s", song->artist, song->title);
+        }
+      else
+        {
+          formatted_name = g_strdup_printf ("%s", song->title);
+        }
+      break;
+
+    case XFMPC_SONG_FORMAT_ALBUM_TITLE:
+      if (NULL != song->album)
+        {
+          formatted_name = g_strdup_printf ("%s - %s", song->album, song->title);
+        }
+      else
+        {
+          formatted_name = g_strdup_printf ("%s", song->title);
+        }
+      break;
+
+    case XFMPC_SONG_FORMAT_ARTIST_TITLE_DATE:
+      if (NULL != song->artist && NULL != song->date)
+        {
+          formatted_name = g_strdup_printf ("%s - %s (%s)", song->artist, song->title, song->date);
+        }
+      else if (NULL != song->date)
+        {
+          formatted_name = g_strdup_printf ("%s (%s)", song->title, song->date);
+        }
+      else if (NULL != song->artist)
+        {
+          formatted_name = g_strdup_printf ("%s - %s", song->artist, song->title);
+        }
+      else
+        {
+          formatted_name = g_strdup_printf ("%s", song->title);
+        }
+      break;
+
+    case XFMPC_SONG_FORMAT_ARTIST_ALBUM_TITLE:
+      if (NULL != song->artist && NULL != song->album)
+        {
+          formatted_name = g_strdup_printf ("%s - %s - %s", song->artist, song->album, song->title);
+        }
+      else if (NULL != song->album)
+        {
+          formatted_name = g_strdup_printf ("%s - %s", song->album, song->title);
+        }
+      else if (NULL != song->artist)
+        {
+          formatted_name = g_strdup_printf ("%s - %s", song->artist, song->title);
+        }
+      else
+        {
+          formatted_name = g_strdup_printf ("%s", song->title);
+        }
+      break;
+
+    case XFMPC_SONG_FORMAT_ARTIST_ALBUM_TRACK_TITLE:
+      formatted_name = g_strdup ("");
+
+      if (song->artist != NULL)
+        {
+          tmp = g_strconcat (formatted_name, song->artist, " - ", NULL);
+          g_free (formatted_name);
+          formatted_name = tmp;
+        }
+
+      if (song->album != NULL)
+        {
+          tmp = g_strconcat (formatted_name, song->album, " - ", NULL);
+          g_free (formatted_name);
+          formatted_name = tmp;
+        }
+
+      if (song->track != NULL)
+        {
+          tmp = g_strconcat (formatted_name, song->track, ". ", NULL);
+          g_free (formatted_name);
+          formatted_name = tmp;
+        }
+
+      tmp = g_strconcat (formatted_name, song->title, NULL);
+      g_free (formatted_name);
+      formatted_name = tmp;
+      break;
+
+    default:
+      g_critical ("Unsupported song format (%d)", song_format);
+      formatted_name = g_path_get_basename (song->file);
+      break;
+    }
+
+  return formatted_name;
+}
+
+static gchar *
+_get_formatted_name_custom (mpd_Song *song,
+                            const gchar *format)
+{
+  gchar *formatted_name, *tmp;
+
+  g_return_val_if_fail (G_LIKELY (format != NULL), NULL);
+
+  formatted_name = g_strdup ("");
+
+  do
+    {
+      if (*format == '%')
+        {
+          format ++;
+
+          switch (*(format))
+            {
+            case 'a':
+              tmp = g_strconcat (formatted_name, song->artist, NULL);
+              break;
+
+            case 'A':
+              tmp = g_strconcat (formatted_name, song->album, NULL);
+              break;
+
+            case 'd':
+              tmp = g_strconcat (formatted_name, song->date, NULL);
+              break;
+
+            case 'D':
+              tmp = g_strconcat (formatted_name, song->disc, NULL);
+              break;
+
+            case 'f':
+              tmp = g_strconcat (formatted_name, song->file, NULL);
+              break;
+
+            case 'g':
+              tmp = g_strconcat (formatted_name, song->genre, NULL);
+              break;
+
+            case 't':
+              tmp = g_strconcat (formatted_name, song->title, NULL);
+              break;
+
+            case 'T':
+              tmp = g_strconcat (formatted_name, song->track, NULL);
+              break;
+
+            default:
+              tmp = g_strconcat (formatted_name, "%", NULL);
+              format --;
+              break;
+            }
+
+          g_free (formatted_name);
+          formatted_name = tmp;
+        }
+      else
+        {
+          tmp = g_strdup_printf ("%s%c", formatted_name, *format);
+          g_free (formatted_name);
+          formatted_name = tmp;
+        }
+    }
+  while (*format ++);
+
+  return formatted_name;
+}
+
+XfmpcSongInfo *
+xfmpc_mpdclient_get_song_info (XfmpcMpdclient *mpdclient, gint id)
+{
+  XfmpcMpdclientPrivate *priv = XFMPC_MPDCLIENT (mpdclient)->priv;
+  XfmpcSongInfo *song_info = g_slice_new0 (XfmpcSongInfo);
+  mpd_Song *song = mpd_playlist_get_song (priv->mi, id);
+
+  song_info->filename = song->file;
+  song_info->artist = song->artist;
+  song_info->title = song->title;
+  song_info->album = song->album;
+  song_info->track = song->track;
+  song_info->date = song->date;
+  song_info->genre = song->genre;
+
+  return song_info;
+}
+
+void
+xfmpc_song_info_free (XfmpcSongInfo *song_info)
+{
+  g_free (song_info->filename);
+  g_free (song_info->artist);
+  g_free (song_info->title);
+  g_free (song_info->album);
+  g_free (song_info->date);
+  g_free (song_info->track);
+  g_free (song_info->genre);
+
+  g_slice_free (XfmpcSongInfo, song_info);
 }
 
